/**
 * Copyright (c) 2000-present Liferay, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of the Liferay Enterprise
 * Subscription License ("License"). You may not use this file except in
 * compliance with the License. You can obtain a copy of the License by
 * contacting Liferay, Inc. See the License for the specific language governing
 * permissions and limitations under the License, including but not limited to
 * distribution rights of the Software.
 *
 *
 *
 */

package com.liferay.poshi.runner.elements;

import com.google.common.reflect.ClassPath;

import com.liferay.poshi.runner.PoshiRunnerContext;
import com.liferay.poshi.runner.util.Dom4JUtil;
import com.liferay.poshi.runner.util.PropsUtil;
import com.liferay.poshi.runner.util.RegexUtil;

import java.io.IOException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.dom4j.Attribute;
import org.dom4j.Comment;
import org.dom4j.Element;
import org.dom4j.Node;
import org.dom4j.tree.DefaultElement;

/**
 * @author Kenji Heigel
 */
public abstract class PoshiElement
	extends DefaultElement implements PoshiNode<Element, PoshiElement> {

	public PoshiElement() {
		super("");
	}

	@Override
	public void add(Attribute attribute) {
		if (attribute instanceof PoshiElementAttribute) {
			super.add(attribute);

			return;
		}

		super.add(new PoshiElementAttribute(attribute));
	}

	public abstract PoshiElement clone(
		PoshiElement parentPoshiElement, String readableSyntax);

	public PoshiElement clone(String readableSyntax) {
		return clone(null, readableSyntax);
	}

	public boolean isReadableSyntaxComment(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		if (readableSyntax.startsWith("//")) {
			return true;
		}

		if (isMultilineReadableSyntaxComment(readableSyntax)) {
			return true;
		}

		return false;
	}

	@Override
	public boolean remove(Attribute attribute) {
		if (attribute instanceof PoshiElementAttribute) {
			return super.remove(attribute);
		}

		for (PoshiElementAttribute poshiElementAttribute :
				toPoshiElementAttributes(attributes())) {

			if (poshiElementAttribute.getAttribute() == attribute) {
				return super.remove(poshiElementAttribute);
			}
		}

		return false;
	}

	@Override
	public String toReadableSyntax() {
		StringBuilder sb = new StringBuilder();

		for (Node node : Dom4JUtil.toNodeList(content())) {
			if (node instanceof PoshiComment) {
				PoshiComment poshiComment = (PoshiComment)node;

				sb.append(poshiComment.toReadableSyntax());
			}
			else if (node instanceof PoshiElement) {
				PoshiElement poshiElement = (PoshiElement)node;

				sb.append(poshiElement.toReadableSyntax());
			}
		}

		return sb.toString();
	}

	protected PoshiElement(String name, Element element) {
		super(name);

		if (!isElementType(name, element)) {
			throw new RuntimeException(
				"Element does not match expected Poshi element name\n" +
					element.toString());
		}

		_addAttributes(element);
		_addNodes(element);
	}

	protected PoshiElement(
		String name, List<Attribute> attributes, List<Node> nodes) {

		super(name);

		if (attributes != null) {
			for (Attribute attribute : attributes) {
				add(attribute);
			}
		}

		if (nodes != null) {
			for (Node node : nodes) {
				add(node);
			}
		}
	}

	protected PoshiElement(
		String name, PoshiElement parentPoshiElement, String readableSyntax) {

		super(name);

		setParent(parentPoshiElement);

		parseReadableSyntax(readableSyntax);

		detach();
	}

	protected String createReadableBlock(String content) {
		StringBuilder sb = new StringBuilder();

		String pad = getPad();

		sb.append("\n");
		sb.append(pad);
		sb.append(getBlockName());
		sb.append(" {");

		if (content.startsWith("\n\n")) {
			content = content.replaceFirst("\n\n", "\n");
		}

		content = content.replaceAll("\n", "\n" + pad);

		sb.append(content.replaceAll("\n\t\n", "\n\n"));

		sb.append("\n");
		sb.append(pad);
		sb.append("}");

		return sb.toString();
	}

	protected abstract String getBlockName();

	protected String getBracedContent(String readableSyntax) {
		return RegexUtil.getGroup(readableSyntax, ".*?\\{(.*)\\}", 1);
	}

	protected String getFileType() {
		PoshiElement poshiParentElement = (PoshiElement)getParent();

		return poshiParentElement.getFileType();
	}

	protected String getNameFromAssignment(String assignment) {
		String name = assignment.split("=")[0];

		name = name.trim();
		name = name.replaceAll("@", "");
		name = name.replaceAll("property ", "");

		return name.replaceAll("var ", "");
	}

	protected String getPad() {
		return "\t";
	}

	protected String getParentheticalContent(String readableSyntax) {
		return RegexUtil.getGroup(readableSyntax, ".*?\\((.*)\\)", 1);
	}

	protected String getQuotedContent(String readableSyntax) {
		return RegexUtil.getGroup(readableSyntax, ".*?\"(.*)\"", 1);
	}

	protected String getReadableCommandKeyword() {
		PoshiElement poshiParentElement = (PoshiElement)getParent();

		return poshiParentElement.getReadableCommandKeyword();
	}

	protected String getReadableEscapedContent(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		return readableSyntax.substring(3, readableSyntax.length() - 3);
	}

	protected String getSingleQuotedContent(String readableSyntax) {
		return RegexUtil.getGroup(readableSyntax, ".*?\'(.*)\'", 1);
	}

	protected String getValueFromAssignment(String assignment) {
		assignment = assignment.trim();

		int start = assignment.indexOf("=");

		int end = assignment.length();

		if (assignment.endsWith(";")) {
			end = end - 1;
		}

		String value = assignment.substring(start + 1, end);

		return value.trim();
	}

	protected boolean isBalancedReadableSyntax(String readableSyntax) {
		readableSyntax = readableSyntax.replaceAll("<!--.*?-->", "");

		readableSyntax = readableSyntax.replaceAll("\'\'\'.*?\'\'\'", "\"\"");

		Stack<Character> stack = new Stack<>();

		for (char c : readableSyntax.toCharArray()) {
			if (!stack.isEmpty()) {
				Character topCodeBoundary = stack.peek();

				if (c == _codeBoundariesMap.get(topCodeBoundary)) {
					stack.pop();

					continue;
				}

				if (topCodeBoundary == '\"') {
					continue;
				}
			}

			if (_codeBoundariesMap.containsKey(c)) {
				stack.push(c);

				continue;
			}

			if (_codeBoundariesMap.containsValue(c)) {
				return false;
			}
		}

		return stack.isEmpty();
	}

	protected boolean isBalanceValidationRequired(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		if (readableSyntax.endsWith(";") || readableSyntax.endsWith("}")) {
			return true;
		}

		return false;
	}

	protected boolean isConditionValidInParent(
		PoshiElement parentPoshiElement) {

		if (parentPoshiElement instanceof AndPoshiElement ||
			parentPoshiElement instanceof ElseIfPoshiElement ||
			parentPoshiElement instanceof IfPoshiElement ||
			parentPoshiElement instanceof NotPoshiElement ||
			parentPoshiElement instanceof OrPoshiElement) {

			return true;
		}

		return false;
	}

	protected boolean isElementType(String name, Element element) {
		if (name.equals(element.getName())) {
			return true;
		}

		return false;
	}

	protected boolean isMacroReturnVar(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		String value = getValueFromAssignment(readableSyntax);

		if (!value.matches("(?s)^\".*\"$") && !value.matches("(?s)^'.*'$") &&
			!isValidFunctionFileName(value) && !isValidUtilClassName(value)) {

			return true;
		}

		return false;
	}

	protected boolean isMultilineReadableSyntaxComment(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		if (readableSyntax.endsWith("*/") && readableSyntax.startsWith("/*")) {
			return true;
		}

		return false;
	}

	protected boolean isValidFunctionFileName(String classCommandName) {
		classCommandName = classCommandName.trim();

		for (String functionFileName : functionFileNames) {
			if (classCommandName.startsWith(functionFileName)) {
				return true;
			}
		}

		return false;
	}

	protected boolean isValidReadableBlock(String readableSyntax) {
		readableSyntax = readableSyntax.trim();

		if (readableSyntax.startsWith("property") ||
			readableSyntax.startsWith("static var") ||
			readableSyntax.startsWith("var")) {

			if (readableSyntax.endsWith("\'\'\';") ||
				readableSyntax.endsWith("\";") ||
				readableSyntax.endsWith(");")) {

				return true;
			}

			return false;
		}

		if (isReadableSyntaxComment(readableSyntax)) {
			return true;
		}

		if (isBalanceValidationRequired(readableSyntax)) {
			return isBalancedReadableSyntax(readableSyntax);
		}

		return false;
	}

	protected boolean isValidUtilClassName(String classCommandName) {
		classCommandName = classCommandName.trim();

		for (String utilClassName : utilClassNames) {
			if (classCommandName.startsWith(utilClassName)) {
				return true;
			}
		}

		return false;
	}

	protected String quoteContent(String content) {
		return "\"" + content + "\"";
	}

	protected List<PoshiElementAttribute> toPoshiElementAttributes(
		List<?> list) {

		if (list == null) {
			return null;
		}

		List<PoshiElementAttribute> poshiElementAttributes = new ArrayList<>(
			list.size());

		for (Object object : list) {
			poshiElementAttributes.add((PoshiElementAttribute)object);
		}

		return poshiElementAttributes;
	}

	protected List<PoshiElement> toPoshiElements(List<?> list) {
		if (list == null) {
			return null;
		}

		List<PoshiElement> poshiElements = new ArrayList<>(list.size());

		for (Object object : list) {
			poshiElements.add((PoshiElement)object);
		}

		return poshiElements;
	}

	protected static final Set<String> functionFileNames = new TreeSet<>();
	protected static final Pattern nestedVarAssignmentPattern = Pattern.compile(
		"(\\w*? = \".*?\"|\\w*? = \'\'\'.*?\'\'\'|\\w*? = .*?\\(.*?\\))" +
			"($|\\s|,)",
		Pattern.DOTALL);
	protected static final Set<String> utilClassNames = new TreeSet<>();

	private void _addAttributes(Element element) {
		for (Attribute attribute :
				Dom4JUtil.toAttributeList(element.attributes())) {

			add(new PoshiElementAttribute((Attribute)attribute.clone()));
		}
	}

	private void _addNodes(Element element) {
		for (Node node : Dom4JUtil.toNodeList(element.content())) {
			if (node instanceof Comment || node instanceof Element) {
				add(PoshiNodeFactory.newPoshiNode(node));
			}
		}
	}

	private static final Map<Character, Character> _codeBoundariesMap =
		new HashMap<>();
	private static final Pattern _namespacedfunctionFileNamePattern =
		Pattern.compile(".*?\\.(.*?)\\.function");

	static {
		_codeBoundariesMap.put('\"', '\"');
		_codeBoundariesMap.put('(', ')');
		_codeBoundariesMap.put('{', '}');
		_codeBoundariesMap.put('[', ']');

		try {
			ClassPath classPath = ClassPath.from(
				PropsUtil.class.getClassLoader());

			for (ClassPath.ClassInfo classInfo :
					classPath.getTopLevelClasses(
						"com.liferay.poshi.runner.util")) {

				utilClassNames.add(classInfo.getName());
				utilClassNames.add(classInfo.getSimpleName());
			}

			utilClassNames.add("selenium");
			utilClassNames.add("TestPropsUtil");
		}
		catch (IOException ioe) {
			throw new RuntimeException(ioe);
		}

		for (String namespacedFunctionFileName :
				PoshiRunnerContext.getFilePathKeys()) {

			Matcher matcher = _namespacedfunctionFileNamePattern.matcher(
				namespacedFunctionFileName);

			if (matcher.find()) {
				functionFileNames.add(matcher.group(1));
			}
		}
	}

}